from ..search_attribute import (
    DictAttribute,
    ProjectionExpressionAttribute)
from ...datatype.expression import BaseExpression

class Projectable(object):
    """A mixin to add the project method"""

    def project(self, *datatypes_or_expressions):
        """return a new Request setup with project attributes

        Adds the ProjectionExpression, ExpressionAttributeNames to the
        request_attributes dict

        Args:
            *datatypes: a list of datatypes. List.index() and Map.key()
                can be passed to project specific nested attributes

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.scan.project(Person.name, Person.email)
            {
                "TableName": "people",
                "ProjectionExpression": "#__name  :_name_lwxqt, #__email  :_email_dvsyj",
                "ExpressionAttributeNames": {
                    "#__name": "name",
                    "#__email": "email"
                }
            }
        """
        for val in datatypes_or_expressions:
            names = {}
            if hasattr(val, 'expression_attribute_name'):
                expr = val
                names[expr.expression_attribute_name] = expr.datatype.column_name
            else:
                expr = BaseExpression('', val, None)
                names[expr.expression_attribute_name] = val.column_name
            self.add_attribute(
                ProjectionExpressionAttribute,
                'ProjectionExpression',
                expr)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
        return self



from .expression import (
    EqualityExpression,
    BeginsWithExpression,
    InExpression)

class DynamoDataType(object):
    """Abstract class for all DataTypes

    A DynamoDataType defines a column on the Model. They should be classlevel attributes
    that are used build models and supplement any sort of Request performed on the table.
    This is the base class for all datatypes and defines many shared expressions used by
    all of the different child classes.
    """

    def __init__(self, default=None, column_name="", condition_type="", mapper_cls=None):
        """Constructor for DynamoDataType

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A Mapper class used to interpret/parse data to/from DynamoDB
            condition_type: A string representing one of the types defined by
                DynamoDB for how the data is stored in the database. While DynamoDB
                supports different datatypes, they are all represented by the following:

                ====  ===============
                ====  ===============
                N     All number types
                S     All string types
                B     ByteBuffer / Binary
                BOOL  Booleans
                SS    A set of Strings
                NS    A set of Numbers
                BS    A set of ByteBuffers
                L     Lists of any datatypes
                M     Maps of key/values
                ====  ===============

        """
        self.default = default
        self.set_column_name(column_name)
        self.condition_type = condition_type
        self.mapper = mapper_cls(self) if mapper_cls else None

    def eq(self, value):
        """Build an expression using the = sign

        Used for performing searching exact matches when querying and scanning. It is also
        used in save requests to update and put values on columns

        Paremeters:
            value: Any value that should be used for an equality check

        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.name.eq("Mom"))
            Person.update.key(Person.email.eq("test@test.com")).set(Person.name.eq("Dad"))
        """
        return EqualityExpression("=", self, value)

    def neq(self, value):
        """Build an expression using the <> sign

        Parameters:
            value: Any value that should be used for an not equal check

        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.name.neq("Mom"))
        """
        return EqualityExpression("<>", self, value)

    def gt(self, value):
        """Build an expression using the > sign

        Parameters:
            value: Any value that should be used for a greater than check

        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.age.gt(20))
        """
        return EqualityExpression(">", self, value)

    def gte(self, value):
        """Build an expression using the >= sign

        Parameters:
            value: Any value that should be used for a greater than or equal check
        
        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.age.gte(21))
        """
        return EqualityExpression(">=", self, value)

    def lt(self, value):
        """Build an expression using the < sign

        
        Parameters:
            value: Any value that should be used for a less than check

        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.age.lt(21))
        """
        return EqualityExpression("<", self, value)

    def lte(self, value):
        """Build an expression using the <= sign

        Parameters:
            value: Any value that should be used for a less than or equal check

        Returns:
            An EqualityExpression

        For example::

            Person.scan.filter(Person.age.lte(20))
        """
        return EqualityExpression("<=", self, value)

    def in_(self, *values):
        """Build an InExpression

        InExpressions can only be used with filter() on the model, it cannot be part of
        a KeyConditionExpression. The will filter for the table for values that match
        exactly any of the values passed in as arguments

        Parameters:
            values: anything value to use to filter the table

        Returns:
            An InExpression

        For example::

           Person.scan.filter(Person.name.in_("Mom", "Dad")) 
        """
        return InExpression(self, values)

    def begins_with(self, value):
        """Build a BeginsWithExpression

        Can be used in Filters or KeyConditionExpressions to create a begins_with
        expression.

        Parameters:
            value: a substring to check if the column begins with

        Returns:
            A BeginsWithExpression

        For example::

            Person.scan.filter(Person.name.begins_with("Mo"))
        """
        return BeginsWithExpression(self, value)

    def set_column_name(self, val):
        """Update the column_name of this instance

        Parameters:
            value: a string for the new column name
        """
        self.column_name = val

    def build(self, val):
        """build the column value based on the val passed in

        building is called automatically by the DynamoDataAttribute
        when the model is initialized. It will use the default value when present if
        the val passed in is None

        Parameters:
            val: A value that will be used to build the attribute on the instance

        Returns:
            The passed in val or the default when the default is set
        """
        if val is None:
            return self._get_default(val)
        return val

    def _get_default(self, val=None):
        """get the default value from the datatype"""
        if self.default:
            if callable(self.default):
                return self.default()
            else:
                return self.default
        return None



from .response import Response

class GetResponse(Response):
    """A Response class to handle GetRequest

    Attributes:
        item: The reconstructed data from the response Item It will be None if the response
            for some reason is missing the ``Item`` key.
    """

    def __init__(self, response, reconstructor):
        """constructor for GetResponse

        Parameters:
            response: a dict from DynamoDB typically from a get_item request
            reconstructor: a Reconstructor object to help interpret data
        """
        super(GetResponse, self).__init__(response, reconstructor)
        try:
            self.item = self.reconstructor.reconstruct(self._raw['Item'])
        except KeyError:
            self.item = None



from copy import copy
from .base_datatype import DynamoDataType
from .mapper import ModelMapper

class ModelMap(DynamoDataType):
    """A class to represent a nested model as a Map

    A ModelMap automatically sets the column_names to use the dot notation when accessing
    the nested attributes. This allows direct reference to nested keys when making
    requests. For example::

        Parent.scan.filter(Parent.child.name.eq('Zac'))
    """

    def __init__(self, model_cls, mapper_cls=ModelMapper, column_name=""):
        """constructor for ModelMap

        Parameters:
            model_cls: the class of the nested Model being used
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the ModelMapper
            column_name: a string defining the name of the column on the table
        """
        super(ModelMap, self).__init__(
            condition_type="M",
            mapper_cls=mapper_cls,
            column_name=column_name)
        self.model_cls = model_cls
        for column in self.model_cls._columns:
            setattr(self, column.column_name, copy(column))

    def set_column_name(self, val):
        """Update the column_name of this instance.

        Automatically apply a dot notation chain from the this column_name to
        all of the nested models datatypes column_names

        Parameters:
            val: the name of the column
        For example::

            class Child(object):
                name = String()

            class Parent(db.Model):
                child = ModelMap(Child)

            Parent.child.name.column_name # <== 'child.name'
        """
        super(ModelMap, self).set_column_name(val)
        for name, attr in self.__dict__.items():
            if isinstance(attr, DynamoDataType):
                new_name = val + "." + name
                attr.set_column_name(new_name)



from ..search_attribute import SearchAttribute

class Returnable(object):
    """A mixin to add the returns method"""

    def returns(self, value):
        """return the Request setup with ReturnValues attribute

        Adds the ReturnValues to the request attributes dict

        Args:
            value: NONE | ALL_OLD | UPDATED_OLD | ALL_NEW | UPDATED_NEW

        Returns:
            the caller of the method. This allows for chaining

        For example::


            from cerami.request.return_values import UPDATED_NEW
            Person.update \\
                .key(Person.email.eq('test@test.com')) \\
                .set(Person.name.eq('new name')) \\
                .returns(UPDATED_NEW) \\
                .build()
            {
                "TableName": "people",
                "ReturnValues": "UPDATED_NEW",
                "Key": {
                    "email": {
                        "S": "test@test.com"
                    }
                },
                "UpdateExpression": "SET #__name = :_name_zhvzz",
                "ExpressionAttributeNames": {
                    "#__name": "name"
                },
                "ExpressionAttributeValues": {
                    ":_name_zhvzz": {
                        "S": "new name"
                    }
                }
            }
        """
        self.add_attribute(
            SearchAttribute,
            'ReturnValues',
            value)
        return self



from .response import Response

class SearchResponse(Response):
    """A Response class to handle SearchRequest

    A SearchResponse is returned from a ScanRequest or QueryRequest

    Attributes:
        count: The number of records returned
        scanned_count: I dont really know how this one works yet
        last_evaluated_key: The primary key of the item where the operation stopped,
            inclusive of the result set. Can be used for pagination for the start value
            in the next query.
        _items: The raw data of all records in the result set
    """

    def __init__(self, response, reconstructor):
        """Constructor for SearchResponse

        Parameters:
            response: a dict from DynamoDB typically from a scan or query request
            reconstructor: a Reconstructor object to help interpret data
        """
        super(SearchResponse, self).__init__(response, reconstructor)
        self.count = self._raw['Count']
        self.scanned_count = self._raw['ScannedCount']
        self.last_evaluated_key = self._raw.get('LastEvaluatedKey')
        self._items = self._raw.get('Items', [])

    @property
    def items(self):
        """A generator to get individual items from the response

        Whenever an item is yielded, it will reconstruct the item using the reconstructor

        Returns:
            A generator to iterate over items in the response

        For example::

            response = Person.scan.filter(Person.age.lt(100)).execute()
            for person in response.items:
                print(person.name)
        """
        for item in self._items:
            yield self.reconstructor.reconstruct(item)



from .base_datatype import DynamoDataType
from .boolean import Boolean
from .base_number import BaseNumber
from .base_string import BaseString
from .byte_buffer import ByteBuffer
from .datetime import Datetime
from .model_map import ModelMap
from .number import Number
from .set import Set
from .string import String
from .uniform_list import UniformList
from .dynamic import (
    Map,
    List,
    DefaultMapGuesser,
    DefaultParseGuesser)



from .base_expression import BaseExpression

class BeginsWithExpression(BaseExpression):
    """A class to generate a `BEGINS WITH` expression for querying/scanning

    When using with a `QueryRequest`, a `BeginsWithExpression` can only be used on the
    sort key. It cannot be used on the partition key

    For example::

        # You can use Person.email.begins_with instead!
        expression = BeginsWithExpression(Person.email, "test")
        Email.scan.filter(expression).build()
        {
            "TableName": "people",
            "FilterExpression": "begins_with(#__email, :_email_xfdww)",
            "ExpressionAttributeNames": {
                "#__email": "email"
            },
            "ExpressionAttributeValues": {
                ":_email_xfdww": {
                    "S": "test@test.com"
                }
            }
        }
    """
    def __init__(self, datatype, value):
        """constructor for BeginsWithExpression

        Parameters:
            datatype: a DynamoDataType that the expression is for
            value: a substring to check the column begins with
        """
        super(BeginsWithExpression, self).__init__('begins_with', datatype, value)

    def __str__(self):
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{expression}({attr_name}, {value})".format(
            attr_name=attr_name,
            expression=self.expression,
            value=self.expression_attribute_value_name)



from .base_datatype import DynamoDataType
from .mapper import ByteMapper

class ByteBuffer(DynamoDataType):
    """A class for all ByteBuffer and Binary datatypes"""

    def __init__(self, mapper_cls=ByteMapper, default=None, column_name=""):
        """constructor for the ByteBuffer

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: a mapper class to manipulate data to/from dynamodb.
                Defaults to the ByteMapper
        """
        super(ByteBuffer, self).__init__(
            mapper_cls=mapper_cls,
            condition_type="B",
            default=default,
            column_name=column_name)



class BaseDatatypeMapper(object):
    """The base class for all Mappers

    Mappers are responsible for translating between a model's data and DynamoDB. All
    requests need to have the data formatted in a specific way, and all responses from
    DynamoDB are formatted similarly. A Mapper is responsible for converting to and from
    this DynamoDB structure.

    Attributes:
        datatype: a DynamoDataType object the mapper is used with
        condition_type: The condition_type of the datatype

    For example::

        mapper = BaseDatatypeMapper(Person.email)
        mapper.map("test@test.com")
        {
            "S": "test@test.com"
        }

        mapper.reconstruct({"S": "test@test.com"})
        "test@test.com"
    """
    def __init__(self, datatype):
        """constructor for the BaseDatatypeMapper

        Parameters:
            datatype: a DynamoDataType object the mapper is used with
        """
        self.datatype = datatype
        self.condition_type = self.datatype.condition_type

    def map(self, value):
        """return the value and its condition_type

        Mapping is done when converting the model into a form
        readable by DynamoDB. Mapping involves two steps.
        First it must return a dict of the value.
        Second it must "resolve" the value so DynamoDB can process it.

        Parameters:
            value: anything that should be converted into a DynamoDB formatted dict

        Returns:
            a dict in a format expected by DynamoDB. Its key is the condition_type of this
            instance and value the return value of ``resolve()``. It will automatically
            return ``{"NULL": True}`` when the value passed in is ``None``

        For example::

            mapper = BaseDatatypeMapper(Person.email)
            mapper.map("test@test.com")
            {
                "S": "test@test.com"
            }
        """
        if value == None:
            return {'NULL': True}
        res = {}
        res[self.condition_type] = self.resolve(value)
        return res

    def resolve(self, value):
        """returns the value resolved for dynamodb

        This method is called by ``map()`` to convert the value itself into a format
        that is expected for DynamoDB. For example, all Number datatypes need to be
        converted to strings in before being submitted.

        Parameters:
            value: anything that should be converted to a DynamoDB dict by ``map()``

        Returns:
            the value as is. Any child class should override this method to convert the
            value as necessary.
        """
        return value

    def reconstruct(self, mapped_dict):
        """return the value from the mapped dict for model instantiation

        DynamoDB returns all attributes as a dict. Reconstructing reads this dict and
        "parses" the value. The return value can be used as the attribute on the Model.
        Reconstructing is only responsible for parsing the data as-is from DynamoDB. The
        process of assigning default values is done by the datatype itself.

        Parameters:
            mapped_dict: A DynamoDB dictionary, whose key is the condition type of this
                instance and value is what needs to be `parsed`. The key can also be
                ``"NULL"`` which represents null values in DynamoDB

        Returns:
            the parsed value from the `mapped_dict`. It will return ``None`` when the key
            of the mapped_dict is ``"NULL"``.

        For example::

            mapper = BaseDatatypeMapper(Person.email)
            mapper.reconstruct({"S": "test@test.com"})
            "test@test.com"
        """
        if mapped_dict.get('NULL') == True:
            return None
        return self.parse(mapped_dict[self.condition_type])

    def parse(self, value):
        """Convert the value from DynamoDB into a format for the Model

        This is the opposite of resolving. So when given the DynamoDB dict
        ``{"S": "Mom"}`` it will return ``"Mom"``

        Parameters:
            value: anything that should be converted from a DynamoDB dict's value
                value should never be None, reconstruct will not pass it here
        Returns:
            the value as is. All child classes should override this method to convert the
            value however necessary.
        """
        return value



from .mixins import BaseRequest, Keyable, Projectable
from ..response import GetResponse

class GetRequest(BaseRequest, Keyable, Projectable):
    """A class to perform the get_item request"""

    def execute(self):
        """perform the get_item request

        Returns:
            a GetResponse object built from the get_item response

        For example::

            Person.get.key(Person.email.eq('test@test.com')).execute()
        """
        response = self.client.get_item(**self.build())
        return GetResponse(response, self.reconstructor)




from .base_reconstructor import BaseReconstructor

class ModelReconstructor(BaseReconstructor):
    """A Reconstructor class to convert the DynamoDB response data back into a Model"""

    def __init__(self, model_cls):
        """constructor for ModelReconstructor

        Parameters:
            model_cls: the class of the Model to convert the data into
        """
        self.model_cls = model_cls

    def reconstruct(self, item_dict):
        """Convert the item_dict into a Model instance

        Iterates over all columns in the model_cls and tries to set the value based on the
        item_dict's keys

        Parameters:
            item_dict: a dict from DynamoDB. It will be in the format like
                ``{"a_column": {"S": "my_value"}}``
        Returns:
            an instance of ``model_cls``
        """
        data = {}
        for column in self.model_cls._columns:
            try:
                val = item_dict[column.column_name]
            except KeyError:
                continue
            data[column.column_name] = column.mapper.reconstruct(val)
        return self.model_cls(**data)



from .base_request import BaseRequest
from .filterable import Filterable
from .keyable import Keyable
from .returnable import Returnable
from .projectable import Projectable
from .limitable import Limitable



from tests.helpers.testbase import TestBase
from cerami.datatype import Boolean
from cerami.datatype.mapper import BooleanMapper

class TestBooleanMapper(TestBase):
    def setUp(self):
        self.dt = Boolean()
        self.mapper = BooleanMapper(self.dt)

    def test_resolve(self):
        """it returns 1 for True 0 for False"""
        assert self.mapper.resolve(True) == 1
        assert self.mapper.resolve(False) == 0

    def test_parse(self):
        """it returns True for truthy values False for falsey values"""
        assert self.mapper.parse(1)
        assert not self.mapper.parse(0)




from .base_datatype_mapper import BaseDatatypeMapper

class UniformListMapper(BaseDatatypeMapper):
    """A Mapper class for UniformLists

    Unlike the ListMapper, the UniformListMapper does not need to guess which datatype
    each item in the array is.

    For example::

        mapper = UniformListMapper(StringMapper(String()))
        mapper.map(["hello", "world"])
        {
            "L": [
                {
                    "S": "hello"
                },
                {
                    "S": "world"
                }
            ]
        }

        mapper.reconstruct({"L": [{"S": "hello"}, {"S": "world"}]})
        ["hello", "world"]
    """

    def __init__(self, mapper):
        """constructor for UniformListMapper

        Parameters:
            mapper: a BaseDatatypeMapper object
        """
        self.mapper = mapper
        self.condition_type = "L"

    def resolve(self, value):
        return [self.mapper.map(i) for i in value]

    def parse(self, value):
        return [self.mapper.reconstruct(i) for i in value]



class Response(object):
    """A BaseClass to handle the response from a Request

    All Request.execute() calls make a request to DynamoDB which returns the records
    and extra meta data about the request. A Response object is responsible for
    interpreting this data

    Attributes:
        reconstructor: a Reconstructor object to help interpret data
        _raw: The entire response objected returned from DynamoDB
    """
    def __init__(self, response, reconstructor):
        """constructor for a Response

        Parameters:
            response: a dict returned from DynamoDB based on a Request.execute() call
            reconstructor: a Reconstructor object to help interpret data
        """
        self._raw = response
        self.reconstructor = reconstructor



from ..response import SearchResponse
from .mixins import BaseRequest, Filterable, Projectable, Limitable

class ScanRequest(BaseRequest, Filterable, Projectable, Limitable):
    """A class to perform the scan request"""

    def execute(self):
        """perform the scan request

        Returns:
            a SearchResponse built from the scan response

        For example::

            Person.scan.filter(Person.name.eq("Mom")).execute()
        """
        response = self.client.scan(**self.build())
        return SearchResponse(response, self.reconstructor)



from .mixins import BaseRequest, Filterable, Keyable, Projectable, Limitable
from ..response import SearchResponse
from .search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute)

class QueryRequest(BaseRequest, Filterable, Projectable, Limitable):
    """ A class to perform the query request"""

    def execute(self):
        """perform the query request

        Returns:
            a SearchResponse built from the query response

        For example::

            Person.query.key(Person.email.eq("test@test.com")).execute()
        """
        response = self.client.query(**self.build())
        return SearchResponse(response, self.reconstructor)

    def index(self, index_name):
        """add IndexName to the search_attributes

        Adds the IndexName to the request_attributes dict

        Parameters:
            index_name: a string of the index to query. It can be a local secondary
                index or global secondary index on the table

        Returns:
            the instance of this class

        For example::

            Person.query.index("MyGlobalIndex).key(Person.name.eq('Mom')).build()
            {
                "TableName": "people",
                "IndexName": "MyGlobalIndex",
                "KeyConditionExpression": "#__name = :_name_lrhve",
                "ExpressionAttributeNames": {
                    "#__name": "name"
                },
                "ExpressionAttributeValues": {
                    ":_name_lrhve": {
                        "S": "Mom"
                    }
                }
            }
        """
        self.add_attribute(SearchAttribute, 'IndexName', index_name)
        return self

    def key(self, *expressions):
        """return a new SearchInterface setup with query attributes

        The QueryRequest does not extend Keyable because it uses
        a different name for the attribute - KeyConditionExpression

        Parameters:
            *expressions: a list of BaseExpressions

        Returns:
            the instance of this class
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(
                QueryExpressionAttribute,
                'KeyConditionExpression',
                expression)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeValues',
                expression.value_dict())
        return self



from mock import patch, call, Mock
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.request.search_attribute import (
    DictAttribute,
    QueryExpressionAttribute)
from cerami.request.mixins import Filterable

class TestModel(Model):
    test = String()

class TestRequest(Filterable):
    pass

class TestFilterable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    def test_filter(self):
        """it adds FilterExpression, ExpresionAttributeNames,
        ExpressionAttributeValues"""
        self.request.add_attribute = Mock()
        expression = TestModel.test.eq('123')
        names = {}
        names[expression.expression_attribute_name] = 'test'

        self.request.filter(expression)
        calls = [
            call(QueryExpressionAttribute,
                 'FilterExpression',
                 expression),
            call(DictAttribute,
                 'ExpressionAttributeNames',
                 names),
            call(DictAttribute,
                 'ExpressionAttributeValues',
                 expression.value_dict())
        ]
        self.request.add_attribute.assert_has_calls(calls)



from .search_attribute import SearchAttribute

class ProjectionExpressionAttribute(SearchAttribute):
    """A class specifically to be used with the Projectable mixin"""

    def __init__(self, value=None):
        """constructor for the SearchAttribute

        Parameters:
            value: it should be an array of BaseExpressions.
        """
        value = value or []
        super(ProjectionExpressionAttribute, self).__init__(value)

    def add(self, expression):
        """Update self.value by appending the new expression

        Parameters:
            expression: a BaseExpression
        """
        self.value.append(expression)

    def build(self):
        """Build a comma separated list of all expressions in value

        ProjectionExpression is a comma separated list of all datatypes that should be 
        returned by  the request. All expressions in the value array are just
        empty expressions containing the Datatype.column_name

        Returns:
            A string of all expressions separated by commas
        """
        return ', '.join(str(expr) for expr in self.value)



from .search_attribute import SearchAttribute

class UpdateExpressionAttribute(SearchAttribute):
    """A SearchAttribute used exclusively for ``UpdateRequest``

    Its value is a dict, whose keys are one the update actions ADD | SET | DELETE | UPDATE
    The value of each corresponding key is an array of expressions. This is so each action
    is unique during the build process and the expressions are comma separated::

        Model.add(Model.a_number(10)).add(Model.other_number(20))
        # becomes
        "ADD a_number 10, other_number 20"
    """

    def __init__(self, value=None):
        """constructor for the SearchAttribute

        Parameters:
            value: it should be a dict whose keys are arrays of expressions
        """
        value = value or {}
        super(UpdateExpressionAttribute, self).__init__(value)

    def add(self, update_action):
        """Update the self.value with the update_action

        It will create the array automatically if the key is missing or will append it

        Parameters:
            update_action: an UpdateAction object
        """
        if self.value.get(update_action.action):
            self.value[update_action.action].append(update_action.expression)
        else:
            self.value[update_action.action] = [update_action.expression]

    def build(self):
        """return all grouped expressions of value

        iterate over all keys and create a comma separated string of all expressions for
        each corresponding value array.

        Returns:
            a string of all expressions
        """
        operations = []
        for action, expressions in self.value.items():
            operation = action + ' ' + ', '.join(str(expr) for expr in expressions)
            operations.append(operation)
        return ' '.join(o for o in operations)



from copy import deepcopy
from .dynamic import List
from .mapper import UniformListMapper

class UniformList(List):
    """A class to represent Lists of entirely the same datatype

    The Base List class uses a Guesser to try and determine the datatype to use for
    the each item in the array. A UniformList does not need to guess, since the
    datatype is consistent and stored as an attribute
    """
    def __init__(self, datatype, default=None, column_name=""):
        """constructor for UniformList

        Parameters:
            datatype: Any DynamoDataType that each item in the List will be
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
        """
        super(UniformList, self).__init__(default=default, column_name=column_name)
        self.datatype = datatype
        self.mapper = UniformListMapper(self.datatype.mapper)

    def index(self, idx):
        """build a duplicate datatype with the _index set to idx

        When forming a Request that involves a specific item in the UniformList,
        that item can be specified using this index() method

        Parameters:
            idx: a number for the index of the desired item in the list

        Returns:
            A copy of the datatype with _index set

        For example::

            MyModel.scan.filter(MyModel.number_list.index(2).eq(100)).execute()
        """
        dt = deepcopy(self.datatype)
        dt._index = idx
        return dt



import dateutil.parser
from datetime import datetime, timezone
from .base_string import BaseString
from .mapper import DatetimeMapper

class Datetime(BaseString):
    """A class to represent all Datetime datatypes"""

    def __init__(self, mapper_cls=DatetimeMapper, default=None, column_name=""):
        """constructor for the Datetime

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the DatetimeMapper
        """
        super(Datetime, self).__init__(
            mapper_cls=mapper_cls,
            default=default,
            column_name=column_name)



import random
import string

class BaseExpression(object):
    """The base class for all expressions.

    Expressions are objects that are generated by datatypes that are used in requests -
    querying, filtering, updating etc.

    Attributes:
        expression: a string representing to define the type of expression. For example,
            it can be `"="` to perform an equality expression
        datatype: a DynamoDataType that the expresion is for
        value: anything that is used in the expression
        expression_attribute_name: DynamoDB has a list of reserved words, but since the
            list is so long and impossible to remember, there are likely collisions with
            a table's columns. For this reason, most DynamoDB requests has an option
            `ExpressionAttributeNames`, which is a dict which makes an arbitrary string
            back to the column name. The `ExpressionAttributeName` must start with a
            `"#"`.

            For example, you may have a column called `"action"` (which is a reserved
            word). An `expression_attribute_name` is automatically generated that prevents
            an error.
        expression_attribute_value_name: DynamoDB requires all of its Expressions to use
            a unique string that gets mapped to its ExpressionAttributeValues dict. This
            string must start with a `":"`. All `expression_attribute_value_names` include
            the column_name of the datatype and a series of random characters to ensure
            there are no collisions when multiple expressions are used in a request

    For example::
   
        expression = BaseExpression("=", Parent.email, "test@test.com")
        expression.expression_attribute_name # <== "#__email"
        expression.expression_attribute_value_name # <== ":_email_xfdww"

        Parent.query.filter(expression).build()
        {
            "TableName": "people",
            "FilterExpression": "#__email = :_email_xfdww",
            "ExpressionAttributeNames": {
                "#__email": "email"
            },
            "ExpressionAttributeValues": {
                ":_email_xfdww": {
                    "S": "test@test.com"
                }
            }
        }
    """
    def __init__(self, expression, datatype, value):
        """constructor for the BaseExpression

        Parameters:
            expression: a string representing to define the type of expression. For
                example, it can be `"="` to perform an equality expression
            datatype: a DynamoDataType that the expresion is for
            value: anything that is used in the expression
        """
        self.expression = expression
        self.datatype = datatype
        self.value = value
        column_name_safe = datatype.column_name.replace('.', '_')
        self.expression_attribute_name = "#__{}".format(column_name_safe)
        self.expression_attribute_value_name = self._generate_variable_name(
            column_name_safe)

    def __str__(self):
        """convert the expression into a string

        it will automatically update the expression attribute name to include the
        index when it is included in the datatype for lists.

        Returns:
            a string of the expression that can be used in Requests
        """
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{attr_name} {expression} {value_name}".format(
            attr_name=attr_name,
            expression=self.expression,
            value_name=self.expression_attribute_value_name)

    def attribute_map(self):
        """return the value and its condition_type by calling a DatatypeMapper

        This is only used by Keyable right now. But i think this should be deprecated
        because its confusing and doesnt really add any value
        """
        return self.datatype.mapper.map(self.value)

    def value_dict(self):
        """return the expected dict for expression-attribute-values

        This is used by many of different requests when building search_attributes. Most
        requests require the `ExpressionAttributeValue` option. This will build that
        corresponding property for this particular expression

        Returns:
            a dict that can be used in ExpressionAttributeValue options

        For example::

            expression = Expression("=", Parent.name, "Mom")
            expression.value_dict() # <== {":_name_abc12":{"S": "Mom"}}
        """
        res = {}
        res[self.expression_attribute_value_name] = self.datatype.mapper.map(self.value)
        return res

    def _generate_variable_name(self, column_name):
        """return a random name for the value"""
        letters = string.ascii_lowercase
        random_letters = ''.join(random.choice(letters) for i in range(5))
        return ":_{}_{}".format(column_name, random_letters)



from ..search_attribute import DictAttribute

class Keyable(object):
    """A mixin to add the key method"""

    def key(self, *expressions):
        """return a new Request setup with the Key attribute

        Adds the Key to the request_attributes dict

        Args:
            *expressions: a list of ``BaseExpressions``

        Returns:
            the caller of the method. This allows for chaning

        For example::
        
            Person.get.key(Person.email.eq('test@test.com')).build()
            {
                "TableName": "people",
                "Key": {
                    "email": {
                        "S": "test@test.com"
                    }
                }
            }
        """
        for expression in expressions:
            key_dict = {}
            key_dict[expression.datatype.column_name] = expression.attribute_map()
            self.add_attribute(DictAttribute, 'Key', key_dict)
        return self



from mock import patch, Mock
from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.mapper import BaseDatatypeMapper

class TestBaseDatatypeMapper(TestBase):
    def setUp(self):
        self.dt = String()
        self.mapper = BaseDatatypeMapper(self.dt)

    def test_map_none(self):
        """it returns the NULL object when value is None"""
        assert self.mapper.map(None) == {'NULL': True}

    def test_map(self):
        """it returns a dict
        with the key the condition_type
        and the value the result of resolve()
        """
        with patch('cerami.datatype.mapper.BaseDatatypeMapper.resolve') as resolve:
            resolve.return_value = "mocked"
            res = self.mapper.map('test')
            assert res == {"S": "mocked"}

    def test_reconstruct_null(self):
        """it returns None when mapped_dict is NULL"""
        assert self.mapper.reconstruct({'NULL': True}) == None

    def test_reconstruct_calls_parse(self):
        """calls parse when the value is not NULL"""
        self.mapper.parse = Mock()
        self.mapper.reconstruct({'S': 'test'})
        self.mapper.parse.assert_called_with('test')

    def test_parse(self):
        """parse returns the value"""
        assert self.mapper.parse('test') == 'test'



from .string_mapper import StringMapper
from .base_datatype_mapper import BaseDatatypeMapper

class NumberMapper(BaseDatatypeMapper):
    """A Mapper class for converting number fields into DynamoDB dictionaries

    At the moment, NumberMappers only support integers. The reconstruction process will
    automatically cast any value as in int.

    For example::

        mapper = NumberMapper(Number())
        mapper.map(30)
        {'N': '30'}

        mapper.reconstruct({'N': '30'})
        30
    """
    def resolve(self, value):
        """convert the number into a string"""
        return str(value)

    def parse(self, value):
        """convert the value into an int"""
        return int(value)




from .base_datatype_mapper import BaseDatatypeMapper

class ByteMapper(BaseDatatypeMapper):
    """A Mapper class for byte encoding data

    This mapper is typically used with the ByteBuffer datatype. This class will
    automatically encode the value passed for both mapping and reconstructing. So, even

    For example::

         mapper = ByteMapper(DynamoDataType(condition_type="B"))
         # You can pass a string if you wanted to for example
         mapper.map("hello world")
         {'B': b'hello world'}

         # But the string will not be returned when reconstructing
         mapper.reconstruct({'B': b'hello world'})
         b'helo world'
    """
    def resolve(self, value):
        """UTF-8 encode the value

        Parameters:
            value: a string to be encoded or an already encoded stream.

        Returns:
            a UTF-8 encoded version of value. If the value fails to be utf-8 encoded, it
            will return the value as-is
        """
        return self._encode(value)

    def parse(self, value):
        """UTF-8 encode the value

        Parameters:
            value: a string to be encoded or an already encoded stream.

        Returns:
            a UTF-8 encoded version of value. If the value fails to be utf-8 encoded, it
            will return the value as-is
        """
        return self._encode(value)

    def _encode(self, value):
        """try to encode it or return the value (meaning its already encoded"""
        try:
            return value.encode('utf-8')
        except AttributeError:
            return value



from .base_expression import BaseExpression

class InExpression(BaseExpression):
    """A class to generate an `IN` expression for filtering

    For example::

        # You can use Person.name.in_ instead!
        expression = InExpression(Person.name, ["Mom", "Dad"])
        Person.scan.filter(expression).build()
        {
            "TableName": "people",
            "FilterExpression": "#__name IN (:_name_qmtxc, :_name_wzhlh)",
            "ExpressionAttributeNames": {
                "#__name": "name"
            },
            "ExpressionAttributeValues": {
                ":_name_qmtxc": {
                    "S": "Mom"
                },
                ":_name_wzhlh": {
                    "S": "Dad"
                }
            }
        }
    """
    def __init__(self, datatype, value):
        """constructor for InExpression

        Parameters:
            datatype: a DynamoDataType that the expression is for
            value: an array of values
        """
        super(InExpression, self).__init__('IN', datatype, value)
        self.expression_attribute_values = self._build_expression_attribute_values()

    def __str__(self):
        value_names = ', '.join([k for k,v in self.value_dict().items()])
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{attr_name} {expression} ({value_names})".format(
            attr_name=attr_name,
            expression=self.expression,
            value_names=value_names)

    def value_dict(self):
        """return the expected dict for expression-attribute-values

        This is used by many of different requests when building search_attributes. Most
        requests require the `ExpressionAttributeValue` option. This will build that
        corresponding property for this particular expression. Since the value
        is an array, this method overrides the ``BaseExpression`` implementation.

        Returns:
            a dict that can be used in ExpressionAttributeValue options
        """
        return self.expression_attribute_values

    def _build_expression_attribute_values(self):
        column_name_safe = self.datatype.column_name.replace('.', '_')
        res = {}
        for v in self.value:
            value_name = self._generate_variable_name(column_name_safe)
            value_mapped = self.datatype.mapper.map(v)
            res[value_name] = value_mapped
        return res



class UpdateAction(object):
    """A class used specificallly for UpdateExpressionAttributes"""

    def __init__(self, action, expression):
        """constructor for UpdateAction

        Parameters:
            action: a string ADD | SET | DELETE | UPDATE
            expression: a BaseExpression
        """
        self.action = action
        self.expression = expression



class BaseReconstructor(object):
    """The base class for all Reconstructors

    Reconstructors are used to interpret data from a DynamoDB response. All
    classes that inherit from BaseReconstructor implement ``reconstruct()`` which
    takes in the DynamoDB data and manipulates it accordingly.

    For example, the ModelReconstructor converts the raw DynamoDB response data back
    into the Model of the corresponding table
    """

    def reconstruct(self, item_dict):
        """The abstract reconstruct method to be implemented by child classes

        Its purpose it to convert the item_dict from DynamoDB

        Parameters:
            item_dict: a dict from DynamoDB. It will be in the format like
                ``{"a_column": {"S": "my_value"}}``

        Raises:
            Exception: Not Implemented
        """
        raise Exception("Not implemented")



from .base_datatype_mapper import BaseDatatypeMapper

class BooleanMapper(BaseDatatypeMapper):
    """A Mapper class for converting fields into DynamoDB booleans

    For example::
    
        mapper = BooleanMapper(Boolean())
        mapper.map(True)
        {'BOOL': 1}

        mapper.reconstruct({'BOOL': 1})
        True
    """

    def resolve(self, value):
        return 1 if bool(value) else 0

    def parse(self, value):
        return bool(value)



from .datatype import DynamoDataType
from .data_attribute import DynamoDataAttribute
from .reconstructor import ModelReconstructor
from .request.return_values import ALL_NEW
from .request import (
    GetRequest,
    ScanRequest,
    PutRequest,
    UpdateRequest,
    DeleteRequest,
    QueryRequest)


class ModelMeta(type):
    """The meta class for all Models

    This meta class is automatically sets the column names on every column attribute
    and defines some class level properties for generating requests

    Attributes:
        _columns: An array for accessing all datatypes defined on the Model
    """
    def __new__(cls, clsname, bases, dct):
        """override the class creation

        Add the _columns property and define the column_name for each datatype
        """
        dct["_columns"] = []
        for name, val in dct.items():
            if isinstance(val, DynamoDataType):
                val.set_column_name(name.lower())
                dct["_columns"].append(val)
        return super(ModelMeta, cls).__new__(cls, clsname, bases, dct)

    @property
    def get(cls):
        """Create a GetRequest for the Model

        Returns:
            a GetRequest object
        """
        return GetRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def scan(cls):
        """Create a ScanRequest for the Model

        Returns:
            a ScanRequest object
        """
        return ScanRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def query(cls):
        """Create a QueryRequest for the Model

        Returns:
            a QueryRequest object
        """
        return QueryRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def put(cls):
        """Create a PutRequest for the Model

        Returns:
            a PutRequest object
        """
        return PutRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

    @property
    def update(cls):
        """Create an `UpdateRequest` for the Model

        Returns:
            an UpdateRequest object
        """
        update = UpdateRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))
        return update.returns(ALL_NEW)


    @property
    def delete(cls):
        """Create a `DeleteRequest` for the Model

        Returns:
            a DeleteRequest object
        """
        return DeleteRequest(
            cls.client,
            tablename=cls.__tablename__,
            reconstructor=ModelReconstructor(cls))

class Model(object, metaclass=ModelMeta):
    """This is the base class for all models.

    It automatically uses the `ModelMeta` to define properties for interacting with
    DynamoDB tables. It should be used to create all classes that need make requests
    to DynamoDB.

    Models consist of a series of DynamoDataTypes that define which columns are to be
    saved in the database. Each datatype is also an instance variable that can be
    manipulated through any methods defined on helper methods.

    For example::

        from cerami.decorators import primary_key
        from cerami.datatype import String

        @primary_key("email")
        class Person(Model):
           __tablename__ = "people"

           email = String()
           name = Sting()

        person = Person(email="test@test.com", name="Mom")
    """
    def __init__(self, **data_kwargs):
        """set all column values from data

        It will set any value not present in data but part of
        the models columns to None

        Parameters:
            **data_kwargs: Each kwarg should be the name of one of the columns with a 
                corresponding value
        """
        data = data_kwargs or {}
        data_keys = data.keys()
        for column in self._columns:
            name = column.column_name
            value = data.get(name, None)
            data_exists = name in data or column.default
            attr = DynamoDataAttribute(column, value, initialized=data_exists)
            setattr(self, name, attr)

    def __getattribute__(self, key, full=False):
        """override __getattribute__

        most of the time we want to call attr.get() when accessing a DynamoDataAttribute.
        The full flag can be used to get the actual object.
        """
        attr = super(Model, self).__getattribute__(key)
        if isinstance(attr, DynamoDataAttribute) and not full:
            return attr.get()
        else:
            return attr

    def __setattr__(self, key, value):
        attr = self._get_full_attribute(key)
        if isinstance(attr, DynamoDataAttribute):
            attr.set(value)
        else:
            super(Model, self).__setattr__(key, value)

    def _get_full_attribute(self, key):
        """Get the class-level DynamoDataType object"""
        try:
            return self.__getattribute__(key, full=True)
        except AttributeError:
            return None

    def as_item(self):
        """return all data values in a format for dynamodb

        Returns:
            a dictionary of the model in a format recognized by DynamoDB

        For example::

            person = Person(email="test@test.com", name="Mom")
            person.as_item()
            {'email': {'S': 'test@test.com'}, 'name': {'S': 'Mom'}}
        """
        item = {}
        for column in self._columns:
            name = column.column_name
            attr = self._get_full_attribute(name)
            item[name] = column.mapper.map(attr.value)
        return item

    def delete(self):
        """delete this record from the database

        Automatically perform a `DeleteRequest` based on the model's primary key

        Returns:
            a DeleteResponse
        """
        deleter = self.__class__.delete
        for column in self._primary_key:
            column_name = column.column_name
            deleter = deleter.key(column.eq(getattr(self.data, column_name)))
        return deleter.execute()

    def put(self):
        """add this record to the database

        Automatically perform a `PutRequest` with the objects data

        Returns:
            a SaveResponse
        """
        putter = self.__class__.put
        putter.item(self.as_item())
        return putter.execute()

    def update(self):
        """update this record in the database

        Automatically perform an `UpdateRequest` based on the models attributes. It will
        only update columns that have been initialized or changed.

        Returns:
            a SaveResponse
        """
        updater = self.__class__.update
        for column in self._primary_key:
            column_name = column.column_name
            updater.key(column.eq(getattr(self, column_name)))
        for column in self._columns:
            column_name = column.column_name
            attr = self._get_full_attribute(column_name)
            if (not column in self._primary_key
                and (attr.initialized or attr._changed)):
                updater.set(column.eq(attr.datatype.mapper.resolve(attr.value)))
        return updater.execute()



from .base_datatype_mapper import BaseDatatypeMapper

class StringMapper(BaseDatatypeMapper):
    """A Mapper class for converting string fields into DynamoDB strings

    For example::

        mapper = StringMapper(String())
        mapper.map("hello world")
        {'S': "hello world"}

        mapper.reconstruct({'S': "hello world"})
        "hello world"
    """
    
    def resolve(self, value):
        return str(value)

    def parse(self, value):
        return str(value)



from .response import Response

class SaveResponse(Response):
    """A Response class to handle saves requests

    SaveResponses are generated through PutRequest and UpdateRequest
    """

    def __init__(self, response, reconstructor):
        """constructor for SaveResponse

        Parameters:
            response: a dict from DynamoDB typicall from put_item or update_item
            reconstructor: a Reconstructor object to help interpret data
        """
        super(SaveResponse, self).__init__(response, reconstructor)
        try:
            self.item = self.reconstructor.reconstruct(self._raw['Attributes'])
        except KeyError:
            self.item = None



from .scan_request_test import *
from .query_request_test import *
from .delete_request_test import *
from .get_request_test import *
from .put_request_test import *
from .update_request_test import *

from .search_attribute import *
from .mixins import *



from .base_reconstructor import BaseReconstructor

class RawReconstructor(BaseReconstructor):
    """A Reconstructor class to return the data back exactly as it was given"""

    def reconstruct(self, item_dict):
        """Return the item_dict

        Parameters:
            item_dict: a dict from DynamoDB. It will be in the format like
                ``{"a_column": {"S": "my_value"}}``

        Returns:
            the item_dict passed in
        """
        return item_dict



from .base_datatype import DynamoDataType
from .mapper.boolean_mapper import BooleanMapper

class Boolean(DynamoDataType):
    """A class for Boolean datatypes"""

    def __init__(self, mapper_cls=BooleanMapper, default=None, column_name=""):
        """constructor for the Boolean datatype

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the BooleanMapper
        """
        super(Boolean, self).__init__(
            mapper_cls=mapper_cls,
            condition_type="BOOL",
            default=default,
            column_name=column_name)



from .base_datatype_mapper import BaseDatatypeMapper

class ModelMapper(BaseDatatypeMapper):
    """A Mapper class for converting Models to DynamoDB dictionaries

    The ModelMapper must be initialized with a ModelMap datatype. It depends on the
    model_cls attribute to know which columns and datatypes to use.

    The benefit of using a ModelMapper over a DictMapper is there is no guess work.
    Because all columns are defined, it automatically knows the datatypes to use.

    For example::

        child = Child(gender="female", age=2)
        mapper = ModelMapper(ModelMap(model_cls=Child))
        mapper.map(child)
        {
            "M": {
                "gender": {
                    "S": "female"
                },
                "age": {
                    "N": "2"
                },
                "name": {
                    "NULL": True,
                },
            }
        }

        mapper.parse({"M": {"name": {"S": "Baby"}}})
        <Child object>
    """

    def resolve(self, value):
        """Convert the model into a DynamoDB formatted dictionary

        Any values column values that are not set on the instance being mapped will
        automatically set the ``{"NULL": True}`` for that column

        Parameters:
            value: the model to be converted

        Returns:
            a dictionary where each key is a column of the model and the value is the
            mapped version of the column value
        """
        res = {}
        model_cls = self.datatype.model_cls
        for column in model_cls._columns:
            column_datatype = getattr(model_cls, column.column_name)
            column_value = getattr(value, column.column_name, None)
            res[column.column_name] = column_datatype.mapper.map(column_value)
        return res

    def parse(self, value):
        """Convert the DynamoDB dict into an instance of the model

        Iterate over all columns on the Model. Use the columns mapper to reconstruct
        the value.

        Parameters:
            value: a dictionary where each key represents a column of the model as a 
                DynamoDB formatted dictionary

        Returns:
            an instance of the model class
        """
        data = {}
        model_cls = self.datatype.model_cls
        for column in model_cls._columns:
            try:
                val = value[column.column_name]
            except KeyError:
                continue
            data[column.column_name] = column.mapper.reconstruct(val)
        return model_cls(**data)



from tests.helpers.testbase import TestBase
from cerami.datatype import Number
from cerami.datatype.mapper import NumberMapper

class TestNumberMapper(TestBase):
    def setUp(self):
        self.dt = Number()
        self.mapper = NumberMapper(self.dt)

    def test_resolve(self):
        """it returns the value as a string"""
        assert self.mapper.resolve(1) == '1'

    def test_parse(self):
        """it returns the value as an int"""
        assert self.mapper.parse('1') == 1



from .search_attribute import SearchAttribute

class DictAttribute(SearchAttribute):
    """A class for representing SearchAttributes as a dict"""

    def __init__(self, value=None):
        """constructor for the SearchAttribute

        Parameters:
        value: it should be a dict, or will default to an empty one
        """
        value = value or {}
        super(DictAttribute, self).__init__(value)

    def add(self, value_dict):
        """Update the self.value with the value_dict
        
        This allows for chaining or the same call to overwrite any keys
        that were previously present with the new values represented in
        value_dict
        """
        self.value.update(value_dict)



from ..search_attribute import (
    DictAttribute,
    QueryExpressionAttribute)

class Filterable(object):
    """A mixin to add the filter method"""

    def filter(self, *expressions):
        """return a new Request setup with filter attributes

        Adds the FilterExpression, ExpressionAttributeNames, and
        ExpressionAttributeValue to the request_attributes dict

        Args:
            *expressions: a list of ``BaseExpressions``

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.scan.filter(Person.name.eq('Zac')).filter(Person.age.lt(50)).build()
            {
                "TableName": "people",
                "FilterExpression": "#__name = :_name_pwmbx and #__age < :_age_twtue",
                "ExpressionAttributeNames": {
                    "#__name": "name",
                    "#__age": "age"
                },
                "ExpressionAttributeValues": {
                    ":_name_pwmbx": {
                        "S": "Zac"
                    },
                    ":_age_twtue": {
                        "N": "50"
                    }
                }
            }
        """
        for expression in expressions:
            names = {}
            names[expression.expression_attribute_name] = expression.datatype.column_name
            self.add_attribute(
                QueryExpressionAttribute,
                'FilterExpression',
                expression)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeNames',
                names)
            self.add_attribute(
                DictAttribute,
                'ExpressionAttributeValues',
                expression.value_dict())
        return self



from .base_datatype import DynamoDataType
from .mapper import StringMapper

class BaseString(DynamoDataType):
    """A Base class for all String datatypes"""

    def __init__(self, mapper_cls=StringMapper, default=None, column_name=""):
        """constructor for the BaseString

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the StringMapper
        """
        super(BaseString, self).__init__(
            mapper_cls=mapper_cls,
            condition_type="S",
            default=default,
            column_name=column_name)



from .mixins import BaseRequest
from .search_attribute import DictAttribute
from ..response import SaveResponse

class PutRequest(BaseRequest):
    """A class to perform the put_item request"""

    def execute(self):
        """perform the put_item request

        Returns:
            a SaveResponse object built from the put_item response

        For example::

            Person.put.item({"email": {"S": "test@test.com"}}).execute()
        """
        response = self.client.put_item(**self.build())
        return SaveResponse(response, self.reconstructor)

    def item(self, item_dict):
        """add the item_dict to the search_attributes

        Adds the Item to the request_attributes dict

        Parameters:
            item_dict: a dict mapped to the DynamoDB format. Typically this is done
                through ``Model.as_item()``

        Returns:
            the caller of the method. This allows for chaining

        For example::

            person = Person(email="person@test.com")
            Person.put.item(person.as_item())
            Person.put.item({"email": {"S": "test@test.com"}).build()
            {
                "TableName": "people",
                "Item": {
                    "email": {
                        "S": "test@test.com"
                    }
                }
            }
        """
        self.add_attribute(DictAttribute, 'Item', item_dict)
        return self



from .base_datatype_mapper import BaseDatatypeMapper

class ListMapper(BaseDatatypeMapper):
    """A Mapper class for Lists

    This mapper is typically used with the ``List`` datatype. DynamoDB requires all items
    of a List to also match the DynamoDB format when making requests. This class uses a
    ``MapGuesser`` and a ``ParseGuesser`` to decide how to resolve and reconstruct each
    key/value pair

    Attributes:
        datatype: a DynamoDataType object the mapper is used with
        condition_type: The condition_type of the datatype
        map_guesser: a MapGuesser object, typically associated with the datatype
        parse_guesser: A ParseGuesser object, typically associated with the datatype

    For example::

        mapper = ListMapper(
            DynamoDataType(condition_type="L"),
            DefaultMapGuesser(),
            DefaultParseGuesser())

        mapper.map(["test@test.com", 123])
        {
            "L": [
                {
                    "S": "test@test.com"
                },
                {
                    "N": "123"
                }
            ]
        }

        mapper.reconstruct({'L': [{'S': 'test@test.com'}, {'N': '123'}]})
        ["test@test.com", 123]
    """

    def __init__(self, datatype, map_guesser, parse_guesser):
        """constructor for the ListMapper

        Parameters:
           datatype: a DynamoDataType object the mapper is used with
           map_guesser: a MapGuesser object, typically associated with the datatype
           parse_guesser: A ParseGuesser object, typically associated with the datatype
        """
        super(ListMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def resolve(self, value):
        """convert the value into a DynamoDB formatted dictionary

        Parameters:
            values: a list

        Returns:
            a list where each item is a DynamoDB formatted dictionary
        """
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.map_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.resolve(val))
        return res

    def parse(self, value):
        """convert the DynamoDB formatted dict into a list

        Parameters:
            values: an array of DynamoDB formatted dictionaries

        Returns:
            a list where each item is reconstructed from the guessed datatype of the
            ParseGuesser
        """
        res = []
        for idx, val in enumerate(value):
            guessed_dt = self.parse_guesser.guess(idx, val)
            res.append(guessed_dt.mapper.reconstruct(val))
        return res



from .base_expression import BaseExpression

class ArithmeticExpression(BaseExpression):
    """An expression class for arithmetic operations.

    This is just a convenience class to logically separate its use from
    `EqualityExpressions`. This class is only used in `Number` datatypes and currently
    only in ``Number.add`` and ``Number.subtract``
    """
    pass



from .mixins import BaseRequest, Keyable, Returnable
from ..response import SaveResponse
from ..datatype.expression import (
    UpdateRemoveExpression,
    EqualityExpression)
from .search_attribute import (
    DictAttribute,
    UpdateAction,
    UpdateExpressionAttribute)


class UpdateRequest(BaseRequest, Keyable, Returnable):
    """A class to perform the update request"""

    def execute(self):
        """perform the update_item request

        Returns:
        a SaveResponse object built from the update_item response

        For example::

            Person.update \\
                .key(Person.email.eq('test@test.com')) \\
                .set(Person.name.eq('Mommy')) \\
                .execute()
        """
        response = self.client.update_item(**self.build())
        return SaveResponse(response, self.reconstructor)

    def set(self, *expressions):
        """Add a SET statement to the UpdateExpression

        The UpdateExpression is a comma separated list of instructions to perform on the
        table. SET will change the value or put the value if it does not exist.

        Parameters:
            *expressions: a list of BaseExpressions. In practice though, the expressions
                should all be EqualityExpressions using '=' (Model.column.eq('...'))

        Returns:
            the caller of the method. This allows for chaining
        """
        for expression in expressions:
            self.update_expression('SET', expression)
        return self

    def remove(self, *datatypes):
        """Add a REMOVE statement to the UpdateExpression

        The UpdateExpression is a comma separated list of instructions to perform on the
        table. REMOVE will delete the specified datatypes from the record. REMOVE can also
        be used to remove elements from a List datatype when used with List.index()

        Parameters:
            *datatypes: a list of DynamoDataType objects

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.update \\
                .key(Person.email.eq("test@test.com")) \\
                .remove(Person.name) \\
                .build()
            {
                "TableName": "people",
                "ReturnValues": "ALL_NEW",
                "Key": {
                    "email": {
                        "S": "test@test.com"
                    }
                },
                "UpdateExpression": "REMOVE #__name  :_name_butcw",
                "ExpressionAttributeNames": {
                    "#__name": "name"
                }
            }
        """
        for datatype in datatypes:
            expression = UpdateRemoveExpression(datatype)
            self.update_expression('REMOVE', expression)
        return self

    def add(self, datatype, value):
        """Add an ADD statement to the UpdateExpression

        **Amazon recommends only using with numbers and sets**

        The UpdateExpression is a comma separated list of instructions to perform on the
        table. ADD can be used to change the value of the Number or add to a Set.

        Parameters:
            datatype: a DynamoDataType object to change
            value: the value to change the datatype
                For Numbers the value should be a number, it can be negative to subtract
                For Sets, the value should be an array

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.update.key(Person.email.eq("test@test.com")).add(Person.age, 10).build()
            {
                "TableName": "people",
                "ReturnValues": "ALL_NEW",
                "Key": {
                    "email": {
                        "S": "test@test.com"
                    }
                },
                "UpdateExpression": "ADD #__age  :_age_dzszf",
                "ExpressionAttributeNames": {
                    "#__age": "age"
                },
                "ExpressionAttributeValues": {
                    ":_age_dzszf": {
                        "N": "10"
                    }
                }
            }
        """
        expression = EqualityExpression('', datatype, value)
        return self.update_expression('ADD', expression)

    def delete(self, datatype, value):
        """Add a DELETE statement to the UpdateExpression


        The UpdateExpression is a comma separated list of instructions to perform on the
        table. DELETE can be used to remove one or more elements from a Set only.

        Parameters:
            datatype: a Set DynamoDatatype object
            value: an array of values to remove from the set.

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.update \\
                .key(Person.email.eq("test@test.com")) \\
                .delete(Person.tags, ["cool", "awesome"]) \\
                .build()
            {
                "TableName": "people",
                "ReturnValues": "ALL_NEW",
                "Key": {
                    "name": {
                        "S": "test@test.com"
                    }
                },
                "UpdateExpression": "DELETE #__tags  :_tags_bttwj",
                "ExpressionAttributeNames": {
                    "#__tags": "tags"
                },
                "ExpressionAttributeValues": {
                    ":_tags_bttwj": {
                        "SS": [
                            "cool",
                            "awesome"
                        ]
                    }
                }
            }
        """
        expression = EqualityExpression('', datatype, value)
        return self.update_expression('DELETE', expression)

    def update_expression(self, action, expression):
        """return a Request setup with the update expression attributes

        Adds the UpdateExpression, ExpressionAttributeNames and ExpressionAttributeValues
        to the request_attributes dict

        Parameters:
            action: ADD | SET | DELETE | UPDATE
            expression: a BaseExpression object

        Returns:
            the caller of the method. This allows for chaining
        """
        name = {}
        name[expression.expression_attribute_name] = expression.datatype.column_name
        self.add_attribute(UpdateExpressionAttribute,
                           'UpdateExpression',
                           UpdateAction(action, expression))
        self.add_attribute(DictAttribute,
                           'ExpressionAttributeNames',
                           name)
        self.add_attribute(DictAttribute,
                           'ExpressionAttributeValues',
                           expression.value_dict())
        return self



from mock import patch, call, Mock
from tests.helpers.testbase import TestBase
import cerami.datatype as dt
from cerami.model import Model
from cerami.datatype import String
from cerami.datatype.expression import EqualityExpression
from cerami.request.search_attribute import (
    SearchAttribute,
    DictAttribute,
    QueryExpressionAttribute,)
from cerami.request.mixins import BaseRequest

class TestModel(Model):
    test = String()

class TestBaseRequest(TestBase):
    def setUp(self):
        self.client = Mock()
        self.request = BaseRequest(self.client, tablename='testtable')

    def test_add_attribute(self):
        """it calls add for the search attribute"""
        with patch("cerami.request.search_attribute.SearchAttribute.add") as add:
            self.request.add_attribute(SearchAttribute, 'Test', 123)
            assert isinstance(
                self.request.request_attributes['Test'],
                SearchAttribute)
            add.assert_called_with(123)



from mock import patch
from tests.helpers.testbase import TestBase
from cerami.datatype import String
from cerami.datatype.expression import InExpression
class TestInExpression(TestBase):
    def setUp(self):
        self.dt = String(column_name="testcol")

    def test__init__(self):
        with patch('cerami.datatype.expression.InExpression._build_expression_attribute_values') as build:
            build.return_value = 'mockedExpressionAttributeValues'
            expression = InExpression(self.dt, ['test'])
            assert expression.expression_attribute_values == 'mockedExpressionAttributeValues'

    def test_value_dict(self):
        """returns expression_attribute_values"""
        with patch('cerami.datatype.expression.InExpression._build_expression_attribute_values') as build:
            build.return_value = 'mockedExpressionAttributeValues'
            expression = InExpression(self.dt, ['test'])
            assert expression.value_dict() == 'mockedExpressionAttributeValues'


    def test__str__one_value(self):
        """it returns the expression without any commas in the values"""
        expression = InExpression(self.dt, ['test'])
        attr_name = expression.expression_attribute_name
        value_name = list(expression.value_dict().keys())[0]
        expected = "{} IN ({})".format(attr_name, value_name)
        assert str(expression) == expected

    def test__str__multiple_values(self):
        """it returns the expressions with comma separated values"""
        expression = InExpression(self.dt, ['test', 'test2'])
        attr_name = expression.expression_attribute_name
        value_names = list(expression.value_dict().keys())
        expected = "{} IN ({}, {})".format(attr_name, value_names[0], value_names[1])
        assert str(expression) == expected



from .base_expression import BaseExpression

class UpdateRemoveExpression(BaseExpression):
    """A class use specifically for ``UpdateRequest.remove()``

    A remove expression does not require  any values since it just needs the column_name.
    There isnt an example worth providing for this class. Remove expressions should really
    only be invoked through ``UpdateRequest.remove()``
    """

    def __init__(self, datatype):
        """Constructor for UpdateRemoveExpression

        Parameters:
            datatype: a DynamoDataType that should be removed from a record
        """
        super(UpdateRemoveExpression, self).__init__('', datatype, None)

    def value_dict(self):
        """there is no value for this expression so return an empty dict

        Normally, the ``value_dict()`` of an expression will return dictionary
        mapping the unique name to the value formatted for DynamoDB. However,
        an UpdateRemoveExpression has no values, which is why this returns an
        empty dict. Because ``UpdateRequest.update_expression()`` automatically
        tries to add the values for any expression, it is important this returns
        an empty dict so it does not break DynamoDB when submitting the request

        Returns:
            an empty dict
        """
        return {}



from .search_attribute import SearchAttribute

class QueryExpressionAttribute(SearchAttribute):
    """A class specifically to be used for QueryRequest.key()"""

    def __init__(self, value=None):
        """constructor for the Search Attribute

        Parameters:
            value: it should be an array of BaseExpressions
        """
        value = value or []
        super(QueryExpressionAttribute, self).__init__(value)

    def add(self, expression):
        """Update self.value by appending the new expression

        Arguments:
            expression: a BaseExpression
        """
        self.value.append(expression)

    def build(self):
        """Build a list of expressions separated by and

        KeyConditionExpression is a string of all different expressions that
        identify the keys of the table. It is the primary keys under most
        situation or can be whatever datatypes identify the index used for the query.
        """
        return ' and '.join(str(expr) for expr in self.value)



from mock import patch, call, Mock, MagicMock
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.request.search_attribute import (
    DictAttribute,
    ProjectionExpressionAttribute)
from cerami.request.mixins import Projectable

class TestModel(Model):
    test = String()

class TestRequest(Projectable):
    pass

class TestProjectable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    @patch('cerami.request.mixins.projectable.BaseExpression')
    def test_project(self, expression_mock):
        """it adds FilterExpression, ExpresionAttributeNames,
        ExpressionAttributeValues"""
        expression_instance = MagicMock()
        expression_instance.expression_attribute_name = 'fake_name'
        expression_mock.return_value = expression_instance

        self.request.add_attribute = Mock()
        names = {'fake_name': 'test'}

        self.request.project(TestModel.test)
        calls = [
            call(ProjectionExpressionAttribute,
                 'ProjectionExpression',
                 expression_instance),
            call(DictAttribute,
                 'ExpressionAttributeNames',
                 names),
        ]
        self.request.add_attribute.assert_has_calls(calls)



from .base_datatype_mapper import BaseDatatypeMapper

class DictMapper(BaseDatatypeMapper):
    """A Mapper class for Dicts

    This mapper is typically used with the ``Map`` datatype. DynamoDB requires all nested
    values to also match their DynamoDB format when making requests. This class uses
    a ``MapGuesser`` and ``ParseGuesser`` to decide how to resolve and 
    reconstruct each key/value pair.

    Attributes:
        datatype: a DynamoDataType object the mapper is used with
        condition_type: The condition_type of the datatype
        map_guesser: a MapGuesser object, typically associated with the datatype
        parse_guesser: A ParseGuesser object, typically associated with the datatype

    For example::

        mapper = DictMapper(
            DynamoDataType(condition_type="M"),
            DefaultMapGuesser(),
            DefaultParseGuesser())

        mapper.map({
            "email": "test@test.com",
            "number": 123,
        })
        {
            "M": {
                "email": {
                    "S": "test@test.com"
                },
                "number": {
                    "N": "123"
                }
            }
        }


        mapper.reconstruct({"M": {"email": {"S": "test@test.com"}}})
        {
            "email": "test@test.com",
        }
    """
    def __init__(self, datatype, map_guesser, parse_guesser):
        """constructor for the DictMapper

        Parameters:
           datatype: a DynamoDataType object the mapper is used with
           map_guesser: a MapGuesser object, typically associated with the datatype
           parse_guesser: A ParseGuesser object, typically associated with the datatype
        """
        super(DictMapper, self).__init__(datatype)
        self.map_guesser = map_guesser
        self.parse_guesser = parse_guesser

    def resolve(self, value):
        """convert the value into a DynamoDB formatted dictionary

        Use the MapGuesser to find the datatype Use the datatype's mapper to resolve
        the value

        Parameters:
            value: a dict

        Returns:
            a dict converted from the argument into a DynamoDB formatted dictionary
        """
        res = {}
        for key, val in value.items():
            guessed_dt = self.map_guesser.guess(key, val)
            res[key] = guessed_dt.mapper.map(val)
        return res

    def parse(self, value):
        """convert a DynamoDB formatted dict into a single value

        Use the ParseGuesser to find the datatype. Use the datatype's reconstructor to
        parse the value

        Parameters:
            value: a DynamoDB formatted dictionary

        Returns:
            a single value of DynamoDB formatted dict. The value is
            reconstructed based on the guessed datatype of the `ParseGuesser`
        """
        res = {}
        for key, nested_dict in value.items():
            guessed_dt = self.parse_guesser.guess(key, nested_dict)
            res[key] = guessed_dt.mapper.reconstruct(nested_dict)
        return res



from .base_datatype_mapper import BaseDatatypeMapper

class SetMapper(BaseDatatypeMapper):
    """A Mapper class for converting a Set into a DynamoDB dictionary

    For example::

        mapper = SetMapper(StringMapper(String()))
        mapper.map(["hello", "world"])
        {'SS': ['hello', 'world']}

        mapper.reconstruct({'SS': ['hello', 'world']})
        ['hello', 'world']
    """
    def __init__(self, mapper):
        """constructor for SetMapper

        Parameters:
           mapper: a MapperObject that this Set decoratar shall wrap 
        """
        self.mapper = mapper
        self.condition_type = self.mapper.datatype.condition_type + "S"

    def resolve(self, value):
        return [self.mapper.resolve(i) for i in value]

    def parse(self, value):
        return [self.mapper.parse(i) for i in value]



from mock import call, Mock
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.request.search_attribute import DictAttribute
from cerami.request.mixins import Keyable

class TestModel(Model):
    test = String()

class TestRequest(Keyable):
    pass

class TestKeyable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    def test_key(self):
        """it adds Key"""
        self.request.add_attribute = Mock()
        expression = TestModel.test.eq('123')
        key_dict = {}
        key_dict[expression.datatype.column_name] = expression.attribute_map()

        self.request.key(expression)
        calls = [
            call(DictAttribute, 'Key', key_dict)
        ]
        self.request.add_attribute.assert_has_calls(calls)



import dateutil.parser
from .base_datatype_mapper import BaseDatatypeMapper
from datetime import datetime, timezone

class DatetimeMapper(BaseDatatypeMapper):
    """A Mapper class for Datetimes

    This mapper is typically used with the ``Datetime`` datatype. This class will
    automatically convert the passed in datetime to UTC as is required by DynamoDB

    https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/DynamoDBMapper.DataTypes.html

    For example::

        import datetime
        mapper = DatetimeMapper(String())
        mapper.map(datetime.datetime.now())
        {'S': '2020-05-22T00:03:46.664845+00:00'}

        mapper.reconstruct({'S': '2020-05-22T00:03:46.664845+00:00'})
        datetime.datetime(2020, 5, 22, 0, 4, 33, 144061, tzinfo=tzutc())
    """

    def resolve(self, value):
        """Convert the datetime into an ISO 8601 string

        Parameters:
            value: a datetime object. Can be timezone aware or unaware

        Returns:
            a ISO 8601 string representation of the value in UTC
        """
        return value.replace(tzinfo=timezone.utc).isoformat()

    def parse(self, value):
        """Convert the datetime string into a datetime object

        Since it uses ``dateutil.parser``, this should handle both
        ISO 8601 strings and timezone agnostic values, however
        if the record was creating using the `resolve()` method, a
        utc timezone should automatically be included

        Parameters:
            value: A string representation of a datetime. Typically the value should be
                an ISO 8601 string

        Returns:
            a datetime object
        """
        return dateutil.parser.parse(value)



from .base_datatype import DynamoDataType
from .mapper import SetMapper

class Set(DynamoDataType):
    """A class to represent a Set

    A set is a unique collection of primitive values. It can not be a collection of other
    Sets, Lists or Maps. They are represented in Cerami as arrays of values at the moment.

    For example::

        class Parent(db.Model):
            friends = Set(String())

        parent = Parent(friends=["zac", "mom", "dad"])
    """
    def __init__(self, datatype, column_name="", default=None):
        """constructor for the Set

        The mapper and condition_type for the set is determined automatically
        by the datatype passed in.

        Parameters:
            datatype: The primitive datatype of each item in the set
            column_name: a string defining the name of the column on the table
            default: a default value for the column. It can be a value or function
        """
        super(Set, self).__init__(column_name=column_name, default=default)
        self.datatype = datatype
        self.mapper = SetMapper(self.datatype.mapper)
        self.condition_type = self.datatype.condition_type + "S"



from mock import patch, call, Mock
from tests.helpers.testbase import TestBase
from cerami.model import Model
from cerami.datatype import String
from cerami.request.search_attribute import SearchAttribute
from cerami.request.mixins import Returnable
from cerami.request.return_values import ALL_NEW

class TestModel(Model):
    test = String()

class TestRequest(Returnable):
    pass

class TestReturnable(TestBase):
    def setUp(self):
        self.request = TestRequest()

    def test_returns(self):
        """it adds ReturnValues"""
        self.request.add_attribute = Mock()
        self.request.returns(ALL_NEW)
        calls = [
            call(SearchAttribute, 'ReturnValues', ALL_NEW)
        ]
        self.request.add_attribute.assert_has_calls(calls)




from copy import copy
from ...reconstructor import RawReconstructor
from ..search_attribute import SearchAttribute

class BaseRequest(object):
    """The Base Class for all requests

    It provides the default constructor and methods for building
    requests.

    Attributes:
        client: A ``boto3.client('dynamodb')``
        request_attributes:a dict of SearchAttribute objects whose
            keys represent options that can be passed to client upon execution
            For example, it may include a FilterExpression key whose value is
            a SearchAttribute that resolves to a string of filters. This is typically
            None but can be used to manually build requests::

                Parent.scan.filter(Parent.name.eq('Mom')).project(Parent.name)
                # The search_attributes can be manually specified
                ScanRequest(client=client, tablename='parents', search_attributes={
                    'FilterExpression': SearchAttribute('name = Mom'),
                    'ProjectionExpression': SearchAttribute('name'),
                })

        reconstructor: a Reconstructor object
    """

    def __init__(self, client, tablename="", request_attributes=None, reconstructor=None):
        """constructor for base request


        Parameters:
            client: A ``boto3.client('dynamodb')``
            tablename: the name of the table to perform the request on
            request_attributes:a dict of SearchAttribute objects whose
                keys represent options that can be passed to client upon execution
                For example, it may include a FilterExpression key whose value is
                a SearchAttribute that resolves to a string of filters. This is typically
                None but can be used to manually build requests::

                    Parent.scan.filter(Parent.name.eq('Mom')).project(Parent.name)
                    # The search_attributes can be manually specified
                    ScanRequest(client=client, tablename='parents', search_attributes={
                        'FilterExpression': SearchAttribute('name = Mom'),
                        'ProjectionExpression': SearchAttribute('name'),
                    })

            reconstructor: a Reconstructor object
        """
        self.request_attributes = copy(request_attributes or {})
        self.client = client
        self.reconstructor = reconstructor or RawReconstructor()
        if tablename:
            self.add_attribute(SearchAttribute, 'TableName', tablename)

    def __str__(self):
        return self.build().__str__()

    def add_attribute(self, attr_class, name, value):
        """add a search attribute to a to the request_attributes dict
        
        All search attributes must be unique keys. When the key already exists, it will
        update that value by calling ``add()``. Depending on the ``attr_class``, this
        can overwrite, append, change the value of the existing search attribute.

        Args:
            attr_class: a SearchAttribute class
            name: the name of the attribute
            value: the value that will be added to the SearchAttribute

        For example::

            Person.scan.limit(10)
            # or ...
            Person.scan.add_attribute(SearchAttribute, 'Limit', 10)
        """
        request_attribute = self.request_attributes.get(name, attr_class())
        request_attribute.add(value)
        self.request_attributes[name] = request_attribute

    def build(self):
        """build the dict used by dynamodb

        Returns:
            a dict whose keys matching the keys of the request_attributes and whose values
            are string versions of each attribute

        For example::

            User.scan.filter(User.email.eq('test@test.com')).build()
            {
                "TableName": "Users",
                "FilterExpression": "#__email = :_email_dpxqm",
                "ExpressionAttributeNames": {
                    "#__email": "email"
                },
                "ExpressionAttributeValues": {
                    ":_email_dpxqm": {
                        "S": "test@test.com"
                    }
                }
            }
        """
        return dict((k, v.build()) for k, v in self.request_attributes.items())



from .base_datatype import DynamoDataType
from .mapper import NumberMapper
from .expression import ArithmeticExpression

class BaseNumber(DynamoDataType):
    """A Base class for all Number datatypes"""

    def __init__(self, mapper_cls=NumberMapper, default=None, column_name=""):
        """constructor for the BaseNumber

        Parameters:
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
            mapper_cls: A mapper class to manipulate data to/from dynamodb.
                Defaults to the NumberMapper
        """
        super(BaseNumber, self).__init__(
            mapper_cls=mapper_cls,
            condition_type="N",
            default=default,
            column_name=column_name)

    def add(self, value):
        """add value to number for an UpdateRequest

        Parameters:
            value: a number to add

        Returns:
            An ArithmeticExpression

        For example::

            Person.update.key(Person.email.eq("test@test.com")).set(Person.age.add(10))
        """
        return ArithmeticExpression('+', self, value)

    def subtract(self, value):
        """subtract value from number for an UpdateRequest

        Parameters:
            value: a number to subtract

        For example::

            Person.update \\
                .key(Person.email.eq("test@test.com")) \\
                .set(Person.age.subtract(10))
        """
        return ArithmeticExpression('-', self, value)



class SearchAttribute(object):
    """The base class for all attributes used for ``BaseRequest.add_attribute()``

    All Request objects have a ``request_attributes`` dict whose keys represent options
    for the request and values are different children of this base class.
    ``Request.build()`` iterates all of these attributes and calls this
    class' build method.

    All Request methods return a reference to the caller, so it is possible to chain
    multiple methods of the same thing together.  For example:: 

        Model.filter(Model.column1.eq(1)).filter(Model.column2.eq(2))

    The search attribute needs to decide how to handle duplicate calls. It may want
    to overwrite the first value, or do something to append them, That logic is
    handled by ``SearchAttribute.add()``

    Attributes:
        value: Anything. It will be up to the ``build()`` method how to convert this value
            into a format suitable for DynamoDb
    """

    def __init__(self, value=None):
        """constructor for the SearchAttribute

        Parameters:
            value: it can by anything - whatever should be associated with the name
        """
        self.value = value

    def add(self, value):
        """setter for value

        All Request methods return a reference to the caller, so it is possible to chain
        multiple methods of the same thing together.
        For example, Model.filter(Model.column1.eq(1)).filter(Model.column2.eq(2))

        The search attribute needs to decide how to handle duplicate calls. It may want
        to overwrite the first value, or do something to append them, That logic is
        handled by SearchAttribute.add()

        Parameters:
            value: A new value to be manipulated by the class
        """
        self.value = value

    def build(self):
        """build the SearchAttribute, which is called

        Request has a request_attributes dict whose keys represent options for the request
        and values are different children of this base class. Request.build() iterates all
        of these attributes and calls this SearchAttribute.build().

        Returns:
            the value
        """
        return self.value



from .base_expression import BaseExpression

class EqualityExpression(BaseExpression):
    """An expression class for performing any equality checks

    They are used when filtering (greater than, less than, equals, etc.) but are also
    used for setting and updating attributes using the equals sign.
    """
    pass



from mock import Mock, patch, call
from tests.helpers.testbase import TestBase
from cerami.response import DeleteResponse
from cerami.request import DeleteRequest
from cerami.request.mixins import Returnable, Keyable

class TestDeleteRequest(TestBase):
    def setUp(self):
        self.mocked_client = Mock()
        self.request  = DeleteRequest(
            tablename="test",
            client=self.mocked_client)

    def test_is_returnable(self):
        """it is returnable"""
        assert isinstance(self.request, Returnable)

    def test_is_keyable(self):
        """it is keyable"""
        assert isinstance(self.request, Keyable)

    def test_execute(self):
        """it calls delete_item with the build
        it returns a DeleteResponse
        """
        self.request.build = Mock()
        expected = {"fake": True}
        self.mocked_client.delete_item.return_value = {'Test': True}
        self.request.build.return_value = expected
        res = self.request.execute()
        self.mocked_client.delete_item.assert_called_with(fake=True)
        assert isinstance(res, DeleteResponse)




from .mixins import BaseRequest, Keyable, Returnable
from ..response import DeleteResponse

class DeleteRequest(BaseRequest, Keyable, Returnable):
    """A class to perform delete_item request"""

    def execute(self):
        """perform the delete_item request

        Returns:
            a DeleteResponse object built from the delete_item response

        For example::

            Person.delete.key(Person.email.eq('test@test.com')).execute()
        """
        response = self.client.delete_item(**self.build())
        return DeleteResponse(response, self.reconstructor)



from .base_expression import BaseExpression

class ListAppendExpression(BaseExpression):
    """A class to generate list_append expressions for `UpdateRequests`

    When using an `UpdateRequest`, a `ListAppendExpression` can be used for any
    `List` datatype.

    For example::

       # you can use Person.toys.append() instead!
       expression = ListAppendExpression(Person.toys, [{"color": "red", "name": "car"}]) 
       Person.update.key(Person.email.eq("test@test.com")).set(expression).build()
        {
            "TableName": "people",
            "ReturnValues": "ALL_NEW",
            "Key": {
                "email": {
                    "S": "test@test.com"
                }
            },
            "UpdateExpression": "SET #__toys = list_append(#__toys, :_toys_ihkiy)",
            "ExpressionAttributeNames": {
                "#__toys": "toys"
            },
            "ExpressionAttributeValues": {
                ":_toys_ihkiy": {
                    "L": [
                        {
                            "M": {
                                "color": {
                                    "S": "red"
                                },
                                "name": {
                                    "S": "car"
                                }
                            }
                        }
                    ]
                }
            }
        }
    """
    def __init__(self, datatype, value):
        """constructor for ListAppendExpression

        Parameters:
            datatype: a DynamoDataType that the expression is for
            value: a an array to append to the existing datatype column
        """
        super(ListAppendExpression, self).__init__( '=', datatype, value)

    def __str__(self):
        attr_name = self.expression_attribute_name
        if hasattr(self.datatype, '_index'):
            attr_name = "{}[{}]".format(attr_name, self.datatype._index)
        return "{attr_name} {expression} list_append({attr_name}, {value_name})".format(
            attr_name=attr_name,
            expression=self.expression,
            value_name=self.expression_attribute_value_name)



import numbers
from copy import deepcopy
from .number import Number
from .string import String
from .set import Set
from .base_datatype import DynamoDataType
from .expression import ListAppendExpression
from .mapper import (
    DictMapper,
    ListMapper)

class DefaultMapGuesser(object):
    """A class to guess what datatype to send to DynamoDB based the attribute value

    A generic map or list type can have anything as its keys and values. There is no way
    to know what the Datatypes should be used for certainty because many of the datatypes
    defined (like Datetimes) all are converted to a String before saved in the table.
    This class will look at the values and decide which Datatype to use.

    This class can be extended to handle your specific use cases. For example, if you know
    that the key "my_datetime" should always return a Datetime datatype, the guess()
    method can be overridden to handle specific cases.
    """

    def guess(self, key, value):
        """guess the datatype from the value

        Guessing for mapping means we are trying to figure out the datatype
        in order to convert it into a dict usable by dynamodb. So this
        guesser will makes its guess based on value directly

        Parameters:
            key: the string of the column_name
            value: the value of the column_name

        Returns:
            A DynamoDataType object
        """
        if isinstance(value, numbers.Number):
            return Number()
        elif isinstance(value, dict):
            return Map()
        elif isinstance(value, list):
            return List()
        else:
            return String()


class DefaultParseGuesser(object):
    """A class to decide what datatype to use toe

    A generic map or list type can have anything as its keys and values. There is no way
    to know what the Datatypes should be used for certainty because many of the datatypes
    defined (like Datetimes) all are converted to a String before saved in the table.
    This class will look at the condition_type to determine what Datatype should be used

    This class can be extended to handle your specific use cases. For example, if you know
    that the key "my_datetime" should always return a Datetime datatype, the guess()
    method can be overridden to handle specific cases.
    """

    def guess(self, key, value):
        """guess the datatype from the key within value

        we are guessing on something like {'M': {'test': {'S': 'hello'}}}
        where key is 'test' and value is {'S': 'hello'}
        this will fetch the inner key ('S') and guess from this value

        Parameters:
            key: the column_name of the attribute
            value: a dict whose key is the condition_type and value is the value
                of the column.

        Returns:
        A DynamoDataType object

        Example:
        guess('my_string', {'S': 'hello'})
        """
        attr_key = list(value)[0]
        if attr_key == "N":
            return Number()
        elif attr_key == "S":
            return String()
        elif attr_key == "SS":
            return Set(String())
        elif attr_key == "NS":
            return Set(Number())
        elif attr_key == "L":
            return List()
        elif attr_key == "M":
            return Map()
        else:
            return String()


class Map(DynamoDataType):
    """A class to represent generic Map datatypes

    A Map is used to store nested data within a record. It is essentially a dict where
    each key/value is stored in the database. Since the values can be anything, a
    MapGuesser and ParseGuesser are used to determine the datatype of these nested
    attributes. These guessers are passed to the constructor, and can be customized to
    return specific datatypes if you know the structure of the nested objects.
    """

    def __init__(
            self,
            map_guesser=None,
            parse_guesser=None,
            default=None,
            column_name=""):
        """constructor for Map

        Parameters:
            map_guesser: an object inheriting from DefaultMapGuesser
                Defaults to the DefaultMapGuesser
            parse_guesser: an object inheriting from DefaultParseGuesser
                Defaults to the DefaultParseGuesser.
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
        """
        super(Map, self).__init__(
            condition_type="M",
            default=default,
            column_name=column_name)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()
        self.mapper = DictMapper(self, self.map_guesser, self.parse_guesser)

    def key(self, datatype, key):
        """build a duplicate datatype with the column_name using a dot notiation

        When forming a Request that involves a specific key from the Map, that item can
        be specified using the key() method.

        Parameters:
            datatype: a DynamoDataType instance representing the nested value
            key: the name of the nested value

        Returns:
            A copy of the datatype with an new keyname

        For example::

            Parent.scan.filter(MyModel.child.key(String(), 'name').eq('Zac'))
        """
        column_name = self.column_name + "." + key
        return type(datatype)(column_name=column_name)


class List(DynamoDataType):
    """A class to represent generic List datatypes

    A List is used to store a collection of values of a varying length. It is essentially
    an array where each item is the value stored in the database. Since values can be
    anything, a MapGueser and ParseGuesser are used to determine the datatype of the
    array's items. These guessers are passed to the constructor, and can be customized to
    return specific datatypes if you know the structure of the array and each item.
    """
    def __init__(
            self,
            map_guesser=None,
            parse_guesser=None,
            default=None,
            column_name=""):
        """constructor for List

        Parameters:
            map_guesser: An object inheriting from DefaultMapGuesser
                Defaults to the DefaultMapGuesser
            parse_guesser: An object inheriting from DefaultParseGuesser
                Defaults to the DefaultParseGuesser.
            default: a default value for the column. It can be a value or function
            column_name: a string defining the name of the column on the table
        """
        super(List, self).__init__(
            condition_type="L",
            default=default,
            column_name=column_name)
        self.map_guesser = map_guesser or DefaultMapGuesser()
        self.parse_guesser = parse_guesser or DefaultParseGuesser()
        self.mapper = ListMapper(self, self.map_guesser, self.parse_guesser)

    def append(self, array):
        """Build an expression to add the array to the end of the existing column data

        It can only be used in SET UpdateExpressions.

        Parameters:
            array: can be a list or single value to be appended. When it is a single
                value, it is automatically put inside its own array, before building the
                expression.

        Returns:
            A ListAppendExpression

        For example::

            Person.update \\
                .key(Person.email.eq("test@test.com"))
                .set(Person.toys.append({"color": "red", "name": "car"})
        """
        if not isinstance(array, list):
            array = [array]
        return ListAppendExpression(self, array)

    def index(self, idx, datatype):
        """build a duplicate datatype with the _index set to idx

        When forming a Request that involves a specific item in the List, that item can
        be specified using this index() method.

        Parameters:
            idx: a number for the index of the desired item in the list
            datatype: a DynamoDataType object representing the item indexed

        Returns:
            A copy of the datatype with _index set

        Raises:
            ValueError: An error when the datatype is not an instance of DynamoDataType

        For example::

            MyModel.scan.filter(MyModel.my_list.index(1, String()).eq('world'))
        """
        if not isinstance(datatype, DynamoDataType):
            raise ValueError("datatype must be an instance of DynamoDataType")
        dt = deepcopy(datatype)
        dt._index = idx
        return dt



from ..search_attribute import SearchAttribute

class Limitable(object):
    """A mixin to add the limit method"""

    def limit(self, limit_number):
        """return a new Request setup with the limit attribute

        Adds the Limit to the request_attributes dict

        Args:
            limit_number: a number representing the maximum items to be returned

        Returns:
            the caller of the method. This allows for chaining

        For example::

            Person.scan.limit(10).build()
            {
                "TableName": "people",
                "Limit": 10
            }
        """
        self.add_attribute(
            SearchAttribute,
            'Limit',
            limit_number)
        return self
